import os
import socket
import subprocess

import boto3
import pyodbc
import typer

from .config import Config
from .handlers import step_function, get_7zip


class BaseRestore:
    def __init__(
            self,
            database: str,
            *,
            s3_filename: str = None,
            do_restore: bool = False,
            hostname: str = socket.gethostname()
    ):
        """Data recovery base class

        The class implements the general logic of data backup recovery:
        searching for the last backup copy in the S3 storage, unzipping and,
        if necessary, restoring instead of the current database

        :param database: database name
        :param s3_filename: using a specified file name in the repository for recovery
        :param hostname: override system hostname
        """
        self.cfg: Config = Config.read()
        self.database = database
        self.specified_s3_filename = s3_filename
        self.hostname = hostname
        self.do_restore = do_restore
        self.s3_filename = self.find_latest_backup() if s3_filename is None else s3_filename

        self._download()
        self._decompress()
        self._restore()
        self._clean()

    # region Hidden "step-by-step" functions

    @step_function('\U0001F6F0 Download archive from repository')
    def _download(self):
        self.download()

    @step_function('\U0001F513 Unzipping')
    def _decompress(self):
        self.decompress()

    @step_function('\U0001F553 Restore database')
    def _restore(self):
        if self.do_restore:
            self.restore()

    @step_function('\U0001F4A5 Garbage cleaning')
    def _clean(self):
        self.clean()

    # endregion

    @property
    def zip_filename(self):
        return os.path.join(self.cfg.temp_path, f'{self.database}.bak.7z')

    @property
    def bak_filename(self):
        return os.path.join(self.cfg.temp_path, f'{self.database}.bak')

    def find_latest_backup(self) -> str:
        prefix = f'{self.hostname}__{self.database}__'
        session = boto3.session.Session()
        s3 = session.client(**self.cfg.s3_connection)

        objects = s3.list_objects_v2(Bucket=self.cfg.bucket_name, Prefix=prefix).get('Contents', [])
        if len(objects) == 0:
            typer.secho('No backups found in the repository', fg=typer.colors.RED)
            raise typer.Exit()

        latest = sorted(objects, key=lambda obj: obj['LastModified'], reverse=True)[0]
        return latest['Key']

    def download(self):
        session = boto3.session.Session()
        s3 = session.client(**self.cfg.s3_connection)
        s3.download_file(
            self.cfg.bucket_name,
            self.s3_filename,
            self.zip_filename
        )

    def decompress(self):
        process = subprocess.Popen([
            get_7zip(),
            'e', '-y',
            f'-p{self.cfg.zip_password}' if len(self.cfg.zip_password) > 0 else '',
            f'-o{self.cfg.temp_path}',
            self.zip_filename
        ], stdout=subprocess.DEVNULL)
        process.wait()

    def get_cursor(self):
        raise NotImplemented

    def restore(self):
        raise NotImplemented

    def clean(self):
        os.remove(self.zip_filename)
        if self.do_restore:
            os.remove(self.bak_filename)
        else:
            typer.echo(f'Database backup file restored in {self.bak_filename}')


class RestoreMSSql(BaseRestore):
    def get_cursor(self):
        connection = pyodbc.connect(self.cfg.mssql_connection_string, autocommit=True)
        return connection.cursor()

    def restore(self):
        cursor = self.get_cursor()
        query = f"""
            RESTORE DATABASE [{self.database}] FROM DISK='{self.bak_filename}' WITH REPLACE;
        """
        cursor.execute(query)
        while cursor.nextset():
            pass


class RestorePostres(BaseRestore):
    def restore(self):
        pg_environ = os.environ.copy()
        pg_environ['PGPASSWORD'] = self.cfg.db_password

        process = subprocess.Popen([
            'pg_restore', '--clean', '--format=custom',
            f'--host={self.cfg.db_host}', f'--port={self.cfg.db_port}', f'--username={self.cfg.db_username}',
            f'--dbname={self.database}', self.bak_filename
        ], stdout=subprocess.DEVNULL, env=pg_environ)
        process.wait()



import os
import socket
import subprocess

import boto3
import pyodbc

from .config import Config
from .handlers import step_function, get_7zip, get_now_timestamp


class BaseBackup:
    def __init__(
            self,
            database: str,
            *,
            hostname: str = socket.gethostname()
    ):
        """Backup data base class

        :param database: database name
        :param hostname: override system hostname
        """
        self.cfg: Config = Config.read()
        self.database = database
        self.hostname = hostname
        self.s3_filename = self._get_s3_filename()

        self._backup()
        self._compress()
        self._upload()
        self._clean()

    # region Hidden "step-by-step" functions

    @step_function(f'\U0001F47D Dump database')
    def _backup(self):
        self.backup()

    @step_function(f'\U0001F510 Compress dump')
    def _compress(self):
        self.compress()

    @step_function('\U0001F4BE Upload data to S3')
    def _upload(self):
        self.upload()

    @step_function(f'\U0001F4A5 Garbage cleaning')
    def _clean(self):
        self.clean()

    # endregion

    def _get_s3_filename(self):
        return f'{self.hostname}__{self.database}__{get_now_timestamp()}.7z'.lower()

    @property
    def bak_filename(self) -> str:
        return os.path.join(self.cfg.temp_path, f'{self.database}.bak')

    @property
    def zip_filename(self) -> str:
        return os.path.join(self.cfg.temp_path, f'{self.database}.bak.7z')

    def get_cursor(self):
        raise NotImplemented

    def backup(self):
        raise NotImplemented

    def compress(self):
        process = subprocess.Popen([
            get_7zip(), 'a',
            self.zip_filename,
            self.bak_filename,
            f'-p{self.cfg.zip_password}' if len(self.cfg.zip_password) > 0 else ''
        ], stdout=subprocess.DEVNULL)
        process.wait()

    def clean(self):
        os.remove(self.zip_filename)
        os.remove(self.bak_filename)

    def upload(self):
        session = boto3.session.Session()
        s3 = session.client(**self.cfg.s3_connection)
        s3.upload_file(
            self.zip_filename,
            self.cfg.bucket_name,
            self.s3_filename
        )


class BackupMSSql(BaseBackup):
    def get_cursor(self):
        connection = pyodbc.connect(self.cfg.mssql_connection_string, autocommit=True)
        return connection.cursor()

    def backup(self):
        cursor = self.get_cursor()
        query = f"BACKUP DATABASE [{self.database}] TO DISK='{self.bak_filename}' WITH NOFORMAT, NOINIT, SKIP, NOREWIND;"
        cursor.execute(query)
        while cursor.nextset():
            pass


class BackupPostgres(BaseBackup):
    def backup(self):
        pg_environ = os.environ.copy()
        pg_environ['PGPASSWORD'] = self.cfg.db_password

        process = subprocess.Popen([
            'pg_dump', '--format=custom',
            f'--host={self.cfg.db_host}', f'--port={self.cfg.db_port}', f'--username={self.cfg.db_username}',
            f'--file={self.bak_filename}', f'{self.database}'
        ], stdout=subprocess.DEVNULL, env=pg_environ)
        process.wait()



from typing import Dict, Type
from typing import List

import typer

from .backup import BaseBackup, BackupMSSql, BackupPostgres
from .config import Config
from .restore import BaseRestore, RestoreMSSql, RestorePostres

app = typer.Typer()

BACKUP_DISPATCHER: Dict[str, Type[BaseBackup]] = {
    'mssql'   : BackupMSSql,
    'postgres': BackupPostgres
}

RESTORE_DISPATCHER: Dict[str, Type[BaseRestore]] = {
    'mssql'   : RestoreMSSql,
    'postgres': RestorePostres
}


@app.command()
def backup(
        databases: List[str] = typer.Argument(..., help='Database list')
):
    for database in databases:
        config = Config.read()
        typer.echo(f'\U000026A1 Process backup {database}')
        BACKUP_DISPATCHER[config.db_type](database)
        typer.echo('')

    typer.echo(f'\U0001F389 That\'s all, folks!')


@app.command()
def restore(
        database: str = typer.Argument(..., help='Database name'),
        do_restore: bool = typer.Option(False, help='Restore database from backup')
):
    config = Config.read()

    if do_restore:
        user_agreement = typer.confirm('Are you sure about recovery? The current data will be changed!')
        if not user_agreement:
            raise typer.Abort()

    typer.echo(f'\U000026A1 Process restore {database}')
    RESTORE_DISPATCHER[config.db_type](database, do_restore=do_restore)

    typer.echo(f'\U0001F389 That\'s all, folks!')


@app.command()
def configure():
    config = Config.read(False)
    config.initialize()
    config.save()


if __name__ == "__main__":
    app()



import datetime
import os
import platform
from pathlib import Path

import typer


def type_done():
    typer.secho('DONE', fg=typer.colors.GREEN)


def type_error():
    typer.secho('ERROR', fg=typer.colors.RED)


def step_function(message: str):
    def upper_wrapper(func):
        def wrapper(*args, **kwargs):
            typer.echo(message + ' ... ', nl=False)
            try:
                result = func(*args, **kwargs)
                type_done()
                return result
            except Exception:
                type_error()
                raise

        return wrapper

    return upper_wrapper


def get_7zip():
    if platform.system() == 'Windows':
        pf_environment_variables = ('ProgramFiles', 'ProgramFiles(x86)', 'ProgramW6432', 'PROGRAMDATA')
        pf_path_list = [Path(os.environ.get(var, '')) for var in pf_environment_variables]
        try:
            pf_path = next(filter(lambda x: x != '', pf_path_list))
            return str(pf_path / '7-Zip' / '7z.exe')
        except StopIteration:
            exit('Could not find path to 7-Zip executable file. Is the archiver really installed?')

    else:
        return '7z'


def get_now_timestamp():
    return datetime.datetime.now().strftime('%Y-%m-%d_%H-%M')



from setuptools import setup

from src.config import VERSION

setup(
    name='ubarec',
    description='Universal backup and recovery using S3 repo',
    url='http://github.com/fgbm/ubarec',
    author='Chmelyuk Vladislav',
    author_email='neimp@yandex.ru',
    license='MIT',
    version=VERSION + '.1',
    packages=['src'],
    install_requires=[
        'typer==0.3.2',
        'pyodbc==4.0.30',
        'colorama==0.4.4',
        'boto3==1.17.4',
        'appdirs==1.4.4'
    ],
    zip_safe=False,
    entry_points={
        'console_scripts': ['ubarec=src.main:app'],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
)



from __future__ import annotations
from enum import Enum

from appdirs import user_config_dir
from pathlib import Path
import json
import typer
import tempfile

APPNAME = 'ubarec'
VERSION = '0.1'
CONFIG_FILE = Path(user_config_dir(APPNAME, False, VERSION)) / 'config.json'


class DatabaseType(str, Enum):
    mssql = 'mssql'
    postgres = 'postgres'


class Config:
    def __init__(
            self,
            service_name: str = 's3',
            endpoint_url: str = 'https://storage.yandexcloud.net',
            region_name: str = 'ru-central1',
            aws_access_key_id: str = '',
            aws_secret_access_key: str = '',
            bucket_name: str = '',
            zip_password: str = '',
            db_type: DatabaseType = DatabaseType.mssql,
            db_host: str = 'localhost',
            db_port: int = 1433,
            db_username: str = '',
            db_password: str = '',
            db_driver: str = '{SQL Server}',
            temp_path: str = tempfile.gettempdir()
    ):
        self.service_name = service_name
        self.endpoint_url = endpoint_url
        self.region_name = region_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.bucket_name = bucket_name
        self.zip_password = zip_password
        self.db_type = db_type
        self.db_host = db_host
        self.db_port = db_port
        self.db_username = db_username
        self.db_password = db_password
        self.db_driver = db_driver
        self.temp_path = temp_path

    @property
    def s3_connection(self):
        return {k: v for k, v in self.__dict__.items() if k in (
            'service_name', 'endpoint_url', 'region_name', 'aws_access_key_id', 'aws_secret_access_key'
        )}

    @property
    def mssql_connection_string(self) -> str:
        return ';'.join([
            f'SERVER={self.db_host}',
            f'PORT={self.db_port}',
            f'UID={self.db_username}',
            f'PWD={self.db_password}',
            f'DRIVER={self.db_driver}'
        ])

    @classmethod
    def read(cls, on_err_init:bool = True) -> Config:
        try:
            with open(CONFIG_FILE, 'r', encoding='utf8') as config_file:
                return cls(**json.load(config_file))
        except FileNotFoundError:
            if on_err_init:
                typer.echo('Configuration file not found, \U00002708 parameters must be entered')
                config = Config()
                config.initialize()
                config.save()
                return config

    def save(self):
        CONFIG_FILE.parent.mkdir(parents=True, exist_ok=True)
        with open(CONFIG_FILE, 'w', encoding='utf8') as config_file:
            json.dump(self.__dict__, config_file, indent=4)

    def initialize(self):
        self.service_name = typer.prompt('Service name', self.service_name)
        self.endpoint_url = typer.prompt('Endpoint URL', self.endpoint_url)
        self.region_name = typer.prompt('Region name', self.region_name)
        self.aws_access_key_id = typer.prompt('S3 accesss id key', hide_input=True)
        self.aws_secret_access_key = typer.prompt('S3 accesss secret key', hide_input=True)
        self.bucket_name = typer.prompt('Bucket name', self.bucket_name)
        self.zip_password = typer.prompt('7z password', hide_input=True)
        self.db_type = typer.prompt('Database type', self.db_type, type=DatabaseType)
        self.db_host = typer.prompt('Database hostname', self.db_host)
        self.db_port = typer.prompt('Database connection port', self.db_port, type=int)
        self.db_username = typer.prompt('Database username', self.db_username)
        self.db_password = typer.prompt('Database password', self.db_password)
        self.db_driver = typer.prompt('Database ODBC driver name', self.db_driver)
        self.temp_path = typer.prompt('Location for temporary files', self.temp_path)
